/*****************************************************************************/
/*FILE: HAL.C	Main Routines	                                             */
/*****************************************************************************
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.
 *****************************************************************************/
/* Originally Created by: Homayoon Akhiani	April,1992		     */
/*****************************************************************************
! REVISION HISTORY:
! Who   When            What
!---------------------------------------------------------------
! HA	15-Jul-1992	Fix a bug in .=100 follow by .=80, Listing and 
                        piped input file
! HA	03-May-1992	Adding Steve Morris idea for listing
! HA	28-Apr-1992	Add -size option
! HA    23-Apr-1992	Include stdlib.h (for portability)    
! HA    xx-Apr-1992     First pass
 *****************************************************************************/
/*****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#ifdef VMS
#include <unixio.h>
#include <file.h>
#else
#ifdef THINK_C
#include <console.h>
#include <unix.h>
#include <fcntl.h>
#else
#include <sys/file.h>
#endif
#endif
#include "hal.h"

extern FILE 	*yyin;			/* Input file */
extern int  	yydebug;		/* BISON debugger flag */
extern char*  	yylval;			/* BISON current token value */
extern B32	R[2];
int		line_number;		/* Source line number */
B32		Opc, 			/* Assembled instruction */
		hpc, pc, 		/* Program Counter */
		ic;			/* Instruction counter: pc/4 */
B32		Sa[2];			/* Start offset address */
B32		Se[2];			/* End address */
B32		Saddr;			/* Start offset address */
FILE		*Os;			/* Output for the Scanner tokens */
FILE	  	*Olist;			/* Output for Listing */
FILE		*Ot;		        /* Output for Symbol table transaction*/
int		Ocode;			/* Output for Machine code */
B32		*m;			/* THE IMAGE */
XREF		*ref;			/* The Xref for the listing */
unsigned int	Maxlw;
unsigned int	Bytes;
int		List;			/* Listing ON/OFF */
int		Kcode;			/* Kill Code generator */
int 		Trace;
avl_tree        *SymTable;		/* AVL data base for symbols*/
char		Image[256];
char		Sline[256];
char 		Infilename[256];	/* Input filename */
char 		Lfilename[256];		/* Listing filename */
/*****************************************************************************/
/* MAIN:                                                                     */
/*****************************************************************************/
main(argc,argv)
int argc;
char **argv;
{
B32 c[2];
int status;
int i;
#ifdef THINK_C
 argc = ccommand(&argv);
#endif
 process_cmdline(argc,argv);
 
 SymTable = avl_init_table(strcmp);

 Maxlw = Bytes >> 2;	/* Bytes/4 */
 m = (B32 *)malloc(Bytes*sizeof(char));
 if(m==NULL)   merror("E1:Can't allocate memory,use -size or get more memory");
 ref = (XREF *)malloc(Bytes*sizeof(XREF));
 if(ref==NULL) merror("E2:Can't allocate memory,use -size or get more memory");
 for(i=0;i<Bytes;i++) ref[i].ln = ref[i].g = 0; 
                     
 hpc=Se[1]=Sa[1];
  pc=Se[0]=Sa[0];
 yyparse();
 rewind(yyin);
 if(Se[1] <= hpc)	/* What is the Max PC we have generated */
  {
   Se[1]=hpc;
   if(Se[0] <= pc) Se[0]=pc;
  };
 SUB64(Se,Sa,c);
 ic = c[0] >> 2;

 if(Trace) fprintf(Ot,"Generation Done\n"); 
 avl_foreach(SymTable,symresolve,AVL_FORWARD);
 if(List) generate_list(Lfilename);

 if(!Kcode)
  {
#ifdef THINK_C  
   Ocode = open(Image,O_CREAT|O_TRUNC|O_WRONLY);
#else
   Ocode = open(Image,O_CREAT|O_TRUNC|O_WRONLY,0664);
#endif
   if(Ocode == -1) perror("%%HAL:Can't open the output file:");
   printf(" Writing %d bytes into %s\n",ic*4,Image);
   status = write(Ocode,(char *)m, (ic*4));
   assert(status != 0);
   if(status == -1) perror("Image Write Problem:");
   close(Ocode);
  }
 else printf("%%HAL: No code has been created!\n");
 fclose(yyin);		/* close the source */
 fclose(Ot);
 fclose(Os);
 if(Olist != NULL) fclose(Olist);
}
/*****************************************************************************/
/* SPC:	switch pc, change the pc to h,l	and update ic			     */
/*****************************************************************************/
spc(h,l)
B32 h,l;
{
 if(Se[1] <= hpc)   /* remember the Max PC */
  {
   Se[1]=hpc;
   if(Se[0] <= pc) Se[0]=pc;
  };
 hpc=h;
  pc=l;
 uic();
}
/*****************************************************************************/
/* UPPERCASE                                                                 */
/* convert a string to uppercase                                             */
/*****************************************************************************/
uppercase(s)
char s[];
{
 int i;
 for(i=0; s[i] != '\0'; i++)
   if(s[i] >= 'a' && s[i] <= 'z')
     s[i] += 'A' - 'a';
}
/*****************************************************************************/
/* Used by Scanner to pass the value of the token to the Parser              */ 
/*****************************************************************************/
it(s)
char *s;
{
 yylval = (char *)malloc((strlen(s)+2) * sizeof(char));
 if(yylval == 0) yyerror("Malloc Error in Lexical Analyzer");
 strcpy(yylval,s);
 fprintf(Os,"token=(%s)",s);
 assert(strlen(Sline) < 255);
 if(strlen(Sline) + strlen(s) >= 255)
  {
   char t1[256];
   strcpy(t1,&Sline[strlen(s)+1]);
   strcpy(Sline,t1);
  };
 strcat(Sline," ");
 strcat(Sline,s);
 assert(strlen(Sline) < 255);
}
/*****************************************************************************/
endline()
{
 fprintf(Os,"%s [ic=%d]\n",Sline,ic);
 if(strlen(Sline)==0)	return;
 if(ic == 0)		return;
 if(ic >= Maxlw)
  {
   printf("%%HAL: The Image is larger than %ld bytes.\n",Bytes);
   printf("%%DIA: Use -size option to increase image size\n");
   exit(1);
  };
 Sline[0] = '\0';
}
/*****************************************************************************/
/*****************************************************************************/
nl()
{
 assert(strlen(Sline) < 255);
 Sline[0] = '\0';
/*
 if(strlen(Sline)!=0) strcat(Sline,"\n                       ");
*/
}
/*****************************************************************************/
/* GENERATE_LIST:							     */
/*****************************************************************************/
generate_list(fn)
char fn[];				/* Listing filename */
{
int i,j,k,flg;
char fline[152];
 if(Lfilename[0] == '\0') Olist = fopen("code.lis","w");
 else			  Olist = fopen(Lfilename,"w");
 if(Olist == NULL) merror("Can't open file for the listing.\n");
 fprintf(Olist,"HAL(Hudson Assembler Linker)");
 fprintf(Olist,"%s Compile date: %s (%s)\n",VERSION,__DATE__,__TIME__);
 fprintf(Olist,"HAL disclaims all warranties with regard to the generated code.\n");
 fprintf(Olist,"Listing for: %s\n",Infilename);
/* fprintf(Olist,"Start=%08X %08X End=%08X %08X\n",Sa[1],Sa[0],hpc,pc);*/
 fprintf(Olist,"Start=%08X %08X End=%08X %08X\n",Sa[1],Sa[0],Se[1],Se[0]);
/****
 for(i=0;i<ic;i++)
  {
   fprintf(Olist,"[%08X] %08X",i*4+Saddr,m[i]);
   fprintf(Olist,"\n");
  };
*****/
 for(i=1,flg=0;fgets(fline,150,yyin) != NULL;i++,flg=0)
  {
   for(j=0;j <= ic;j++)
    {
     if(ref[j].g != 0 && ref[j].ln == i)
      {
       flg=1;
       fprintf(Olist,"[%08X] %08X ",j*4+Saddr,m[j]);
       if(ref[j+1].ln <= i && ref[j+1].g != 0) fprintf(Olist,"\n");
      };
    };
		       /* "[12345678] 12345678 "*/
   if(!flg) fprintf(Olist,"                    ");
   fprintf(Olist,"%3d: %s",i,fline);
  };

 fprintf(Olist,"**********************************************************\n");
 fprintf(Olist,"**Symbol Table *******************************************\n");
 fprintf(Olist,"**********************************************************\n");
 avl_foreach(SymTable,symtableit,AVL_FORWARD);
}
/*****************************************************************************/
/*****************************************************************************/
merror(s)
char s[];
{
 printf("\n %s\n",s);
 exit(1);
}
/*****************************************************************************/
/*****************************************************************************/
scan64(s,e)
char *s;
B32  e[];
{
int  len;
char c[15];
 len = strlen(s);
 uppercase(s);
 e[0]=e[1]=0x0;
 if(len > 10)
  {
   sscanf(&s[len-8],"%x",&e[0]);
   strncpy(c,s,(len-8));
   c[len-8]='\0';
   sscanf(c,"^X%x",&e[1]);
/**printf("64bit:[%s]--> %08X %08X\n",s,R[1],R[0]);  **/
  }
 else
  {
   sscanf(s,"^X%x",&e[0]);
  };
 return;
}
/*****************************************************************************/
/* PROCESS_CMDLINE:                                                          */
/* OUTPUT: 								     */
/*****************************************************************************/
process_cmdline(n,arg)
int  n;
char *arg[];
{
int  i;
 printf("HAL(Hudson Assembler Linker)\n");
 printf("%s Compile date: %s (%s)\n",VERSION,__DATE__,__TIME__);
 printf(" %s\n",arg[0]);
 Lfilename[0]= '\0';
 Sline[0]    = '\0';
 line_number =1;
 ic = Kcode = Trace = Saddr = Sa[1] = Sa[0] = 0;
 Bytes = MSIZE;

 if(n == 1 || !strcmp(arg[1],"?") || !strcmp(arg[1],"-h"))
  {
   printf(" hal <inputfile> <outputfile> [-list][-size #][-system #][-verbose]\n");
   printf("\t\t-system ^xaddress: Start address on the code\n");
   printf("\t\t-list filename   : Turn on the listing\n");
   printf("\t\t-size number     : Maximum size for the generated code(Bytes)\n");
   printf("\t\t-verbose         :\n\n");
   printf("\t\t-trace           : trace on\n\n");
   printf("\t\t Start address is a 64 bit long word align address(in Hex).\n");
   printf("\t\t The default start address is ^x0\n");
   printf("\t\t The default size is %ld bytes\n",Bytes);
   exit(1);
  };
 if(n < 3 && arg[1][0] != '-')
  {
   printf(" Not enough parameters: <input file>  <output file>");
   exit(1);
  };

 if(arg[1][0] == '-')
  {
   int c;
   /*yyin    = stdin;*/
   yyin = tmpfile();
   while((c = getc(stdin)) != EOF) putc(c,yyin); 
   rewind(yyin);
  }
 else
   yyin = fopen(arg[1],"r"); 
/*
 if(argc>0)     yyin    = fopen(argv[1],"r");
 else           yyin    = stdin;
*/
 if(yyin == NULL) perror("%%HAL:Can't open the input file:");
 strcpy(Infilename,arg[1]);
 strcpy(Image,arg[2]);
 if ( n > 3 )                      
  {                                   
   for (i = 3; i < n; i++)         
    {                                 
     uppercase(arg[i]);
     if(strncmp(arg[i],"-TR",3)    == 0) Trace=1;
     if(strncmp(arg[i],"-VE",3)    == 0) yydebug=1;
     if(strncmp(arg[i],"-LI",3)    == 0) 
      {
       List = TRUE;
       sscanf(arg[i+1],"%s",Lfilename);
       /*i++; We don't really need this until Options look like filenames*/
      };
     if(strncmp(arg[i],"-SI",3)    == 0) sscanf(arg[i+1],"%ld",&Bytes); 
     if(strncmp(arg[i],"-SYS",4)   == 0) 
      {
/***   sscanf(arg[i+1],"%x",&Saddr); ***/
       scan64h(arg[i+1]);
       Saddr=R[0];
       hpc  =R[1];
       Sa[1]=R[1];Sa[0]=R[0];
      };
    };
  };

 if(Trace)
  {
   Os = fopen("scan.log","w");
   Ot = fopen("tran.log","w");
  }
 else	/* Try to generate the logs into NULL device */
  {
#ifdef VMS
   Os = fopen("NL:","w");
   Ot = fopen("NL:","w");
#else
   Os = fopen("/dev/null","w");
   Ot = fopen("/dev/null","w");
#endif
  };
}


