/*****************************************************************************/
/*FILE: HAL.H - HAL Header file						     */
/*****************************************************************************
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.
 *****************************************************************************/
/* Originally Created by: Homayoon Akhiani	April,1992		     */
/*****************************************************************************
! REVISION HISTORY:
! Who   When            What
!---------------------------------------------------------------
! HA	16-Jul-1992	HW_LD,HW_ST TMOD and THWM constants
! HA	13-May-1992	CASE_SENSITIVE
! HA    xx-Apr-1992     First pass
 *****************************************************************************/
/*****************************************************************************/
#ifndef _HAL_H_
#define _HAL_H_

#include "avl.h"
#define VERSION	"Ver. 0.19"
#define CASE_SENSITIVE	0	    /* 0 means not sensitive to case */
#define MSIZE	(10000*4)	    /* This is the default size of Image */
#define SYSZ	50		    /* Size of the symbols */
#define INUM	100

#ifndef TRUE
#define TRUE	(1)
#define FALSE	(0)
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE	1
#endif
/* Tags for the modifications*/
#define	TD32	1
#define	TD64	2
#define	TMDS	3
#define	TBDS	4
#define	TIOP	5
#define	TMHW	6
#define	TMOD	7

/* Operations */
#define	NONE	0
#define	OADD	1
#define	OSUB	2         
#define	OAND	3
#define	OORR	4
#define	OXOR	5
#define	OONE	6	/* Direct assignment */

typedef unsigned int B32;

union squad {
    B32			ulong[2];
    unsigned char       ubyte[8];
 };
typedef union squad B64;

typedef struct refnode RNODE;
struct refnode {
	unsigned int *pref;		/*Pointer to 32 bit instruction????*/
	int	     ic;		/*index to memory array        ^^^^*/
	char	     instr[80];		/*Instruction in text */
	int	     line;		/*line number */
	int	     type;		/*format of the instruction (Tags)*/
	RNODE	     *next;		/*->next reference */
};
typedef struct symboltable  SENTRY;
struct symboltable {
        char	     symbol[SYSZ];	/*Symbol name */
	int	     valid;		/*value is valid */
	int	     pc;		/* PC that label was defined */
	unsigned int value[2];		/*value of the symbol*/
	int	     operation;
        char	     s1[SYSZ];		/*LH operator */
        char	     s2[SYSZ];		/*RH operator */
	RNODE	     *ref;		/*->refnode s */
};

struct xref_list {
	int	g;	/* Generate instruction */
	B32	ln;	/* line number from source */
	B32	ic;	/* Instruction counter for above line */
};
typedef struct xref_list XREF;
/*****************************************************************************/
/* define the IPR's							     */
/*****************************************************************************/
#define MAX_IPR			(256*2)
#define MAX_ADU_IPR		(128)

#define	IPR_IBOX		(1<<5)
#define	IPR_ABOX		(1<<6)
#define	IPR_PAL			(1<<7)

#define	IPR_DTB_CTL		(IPR_ABOX | 0)
#define	IPR_DTB_PTE		(IPR_ABOX | 2)
#define	IPR_DPTE_TEMP	(IPR_ABOX | 3)
#define	IPR_MM_CSR		(IPR_ABOX | 4)
#define	IPR_VA		(IPR_ABOX | 5)

#define	IPR_ITBZAP		(IPR_IBOX | 6)
#define	IPR_DTBZAP		(IPR_ABOX | 6)
#define	IPR_ITBASM		(IPR_IBOX | 7)
#define	IPR_DTBASM		(IPR_ABOX | 7)
#define	IPR_ITBIS		(IPR_IBOX | 8)
#define	IPR_DTBIS		(IPR_ABOX | 8)

#define	IPR_BIU_ADDR	(IPR_ABOX | 9)
#define	IPR_BIU_STAT	(IPR_ABOX | 10)
#define	IPR_DC_ADDR		(IPR_ABOX | 11)
#define	IPR_DC_STAT		(IPR_ABOX | 12)
#define	IPR_FILL_ADDR	(IPR_ABOX | 13)	
#define	IPR_ABOX_CTL	(IPR_ABOX | 14)
#define	IPR_ALT_MODE	(IPR_ABOX | 15)
#define	IPR_CC		(IPR_ABOX | 16)
#define	IPR_CC_CTL		(IPR_ABOX | 17)
#define	IPR_BIU_CTL		(IPR_ABOX | 18)
#define	IPR_FILL_SYNDROME	(IPR_ABOX | 19)
#define	IPR_BC_TAG		(IPR_ABOX | 20)
#define	IPR_FLUSH_IC	(IPR_ABOX | 21)

#define	IPR_INTR_FLAG	(IPR_ABOX | 22)
#define	IPR_LOCK_FLAG	(IPR_ABOX | 23)

#define	IPR_TB_TAG		(IPR_IBOX | 0)
#define	IPR_ITB_PTE		(IPR_IBOX | 1)
#define	IPR_ICCSR		(IPR_IBOX | 2)
#define	IPR_IPTE_TEMP	(IPR_IBOX | 3)
#define	IPR_EXC_ADDR	(IPR_IBOX | 4)
#define	IPR_SL_RCV		(IPR_IBOX | 5)

#define	IPR_PS		(IPR_IBOX | 9)
#define	IPR_EXC_SUM		(IPR_IBOX | 10)
#define	IPR_PAL_BASE	(IPR_IBOX | 11)
#define	IPR_HIRR		(IPR_IBOX | 12)
#define	IPR_SIRR		(IPR_IBOX | 13)
#define	IPR_ASTRR		(IPR_IBOX | 14)
#define	IPR_HIER		(IPR_IBOX | 16)
#define	IPR_SIER		(IPR_IBOX | 17)
#define	IPR_ASTER		(IPR_IBOX | 18)
#define	IPR_SL_CLR		(IPR_IBOX | 19)
#define	IPR_SL_XMIT		(IPR_IBOX | 22)

#define	IPR_ISSUE_CHK	(IPR_IBOX | 29)
#define	IPR_SINGLE_ISSUE	(IPR_IBOX | 30)
#define	IPR_DUAL_ISSUE	(IPR_IBOX | 31)

#define	IPR_PAL_R0		(IPR_PAL  | 0)
#define	IPR_PAL_R1		(IPR_PAL  | 1)
#define	IPR_PAL_R2		(IPR_PAL  | 2)
#define	IPR_PAL_R3		(IPR_PAL  | 3)
#define	IPR_PAL_R4		(IPR_PAL  | 4)
#define	IPR_PAL_R5		(IPR_PAL  | 5)
#define	IPR_PAL_R6		(IPR_PAL  | 6)
#define	IPR_PAL_R7		(IPR_PAL  | 7)
#define	IPR_PAL_R8		(IPR_PAL  | 8)
#define	IPR_PAL_R9		(IPR_PAL  | 9)
#define	IPR_PAL_R10		(IPR_PAL  | 10)
#define	IPR_PAL_R11		(IPR_PAL  | 11)
#define	IPR_PAL_R12		(IPR_PAL  | 12)
#define	IPR_PAL_R13		(IPR_PAL  | 13)
#define	IPR_PAL_R14		(IPR_PAL  | 14)
#define	IPR_PAL_R15		(IPR_PAL  | 15)
#define	IPR_PAL_R16		(IPR_PAL  | 16)
#define	IPR_PAL_R17		(IPR_PAL  | 17)
#define	IPR_PAL_R18		(IPR_PAL  | 18)
#define	IPR_PAL_R19		(IPR_PAL  | 19)
#define	IPR_PAL_R20		(IPR_PAL  | 20)
#define	IPR_PAL_R21		(IPR_PAL  | 21)
#define	IPR_PAL_R22		(IPR_PAL  | 22)
#define	IPR_PAL_R23		(IPR_PAL  | 23)
#define	IPR_PAL_R24		(IPR_PAL  | 24)
#define	IPR_PAL_R25		(IPR_PAL  | 25)
#define	IPR_PAL_R26		(IPR_PAL  | 26)
#define	IPR_PAL_R27		(IPR_PAL  | 27)
#define	IPR_PAL_R28		(IPR_PAL  | 28)
#define	IPR_PAL_R29		(IPR_PAL  | 29)
#define	IPR_PAL_R30		(IPR_PAL  | 30)
#define	IPR_PAL_R31		(IPR_PAL  | 31)

void		symresolve();
void		symtableit();
int		symget();

/* Header Protection */
#endif			
