%{
/*****************************************************************************/
/*FILE: HAL_PARSER.Y	Grammer Analyzer (use with BISON or YACC)	     */
/*****************************************************************************
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.
 *****************************************************************************/
/* Originally Created by: Homayoon Akhiani	April,1992		     */
/*****************************************************************************
! REVISION HISTORY:
! Who   When            What
!---------------------------------------------------------------
! HA	16-Jul-1992	HW_LD,HW_ST
! HA	20-May-1992	Fix the bug in .LONG, add .ALIGN x, fill value
! HA	01-May-1992	Add more misc. RPCC, FETCH,...
! HA    23-Apr-1992	Include stdlib.h (for portability)    
! HA	17-Apr-1992	.ALIGN mask corrected
! HA    xx-Apr-1992     First pass
 *****************************************************************************/
/*****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "hal.h"
#define alloca  malloc
#define YYSTYPE char*
#define YYDEBUG 1
YYSTYPE	add();
extern XREF *ref;
extern char Sline[];
extern int  Kcode;
extern int  line_number;	/* Source line number */
extern FILE *yyin;		/* Input stream */
extern FILE *Ot;
extern B32  Opc, hpc, pc, ic;
extern B32  *m;
extern B32  Sa[];

B32	    R[2];
B32    	    ev[2],o1[2],o2[2];	/* Experation evaluation parameters*/
char	    Sym[SYSZ];
char	    S1[SYSZ],
	    S2[SYSZ],
	    S3[SYSZ];		/* Hold the symbol name */		
int	    li;			/* index for the label*/
int	    Reg, 
	    sop,		/* Symbolic operation */
	    st;			/* Status of the last symget */
%}
%token EOSC
%token DOT			/* . */
%token CL			/* : */
%token CA			/* , */
%token RP			/* ) */
%token LP			/* ( */
%token PD			/* # */
%token NL			/* \n*/
%token OR			/* ! */
%token XOR			/* \ */
%token REG
%token INS_PAL
%token INS_MEM
%token INS_HWM			/* HW_LD, HW_ST */
%token INS_BRN
%token INS_FOP
%token INS_FP2			/* Floating convert */
%token INS_IOP
%token INS_MIS			/* Misc.(No operand)MB, Draint etc...,HW_REI */
%token INS_M1A			/* Misc.(with Ra)RPCC */
%token INS_M1B			/* Misc.(with Rb)FETCH, FETCH_M */
%token IMMED
%token IM64
%token ALIGN
%token QUAD
%token LONG
%token ADDR
%token BLKL
%token BLKQ
%token ENDP
%token SYMB
%token IPRS
%token COMMENT
%left  '-' '+'			/* same precedence left to right */
%%
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/* Class Description 							     */
/*****************************************************************************/
Lines:
   Line
 | Lines Line                                    
 ;
Line:
   /*empty*/
 | COMMENT			{rs();}
 | Directive			{rs();}
 | Instruction			{rs();}
 | Assigment			{line_number++;rs();}
 | SYMB CL CL			{symdef(Sym,hpc,pc,NULL,NULL,NONE);rs();}
 | SYMB    CL			{symdef(Sym,hpc,pc,NULL,NULL,NONE);rs();}
 | ENDP				{fprintf(Ot,"EoPROGRAM\n"); endline();return;}
 | NL				{line_number++; nl();rs();}
 | error			{yyerror("Unknown statement");}
 ;

Instruction:
   INS_FOP ThreeFReg		{dbug();m[ic++]=Opc;ipc(4);}
 | INS_FP2 TwoReg		{dbug();m[ic++]=Opc;ipc(4);}
 | INS_IOP REG			{Opc|=(Reg<<21);}
	   CA IReg		{dbug();m[ic++]=Opc;ipc(4);}
 | INS_MEM REG			{Opc|=(Reg<<21);}
           CA Disp16 LP REG RP  {dbug();m[ic++]=Opc|(Reg<<16);ipc(4);}
 | INS_HWM REG			{Opc|=(Reg<<21);}
           CA Disp12 LP REG RP Mode {dbug();m[ic++]=Opc|(Reg<<16);ipc(4);}
 | INS_BRN REG CA Disp21	{dbug();m[ic++]=Opc|(Reg<<21);ipc(4);} 
 | INS_PAL REG CA Ipr		{dbug();m[ic++]=Opc|(Reg<<21)|(Reg<<16);ipc(4);}
 | INS_M1A REG			{dbug();m[ic++]=Opc|(Reg<<21);ipc(4);}
 | INS_M1B LP REG RP		{dbug();m[ic++]=Opc|(Reg<<16);ipc(4);}
 | INS_MIS			{dbug();m[ic++]=Opc;ipc(4);}
 | error			{yyerror("Unknown instruction");}
 ;
Directive:
   BLKL    Oper1		{ipc(o1[0]);   uic();}
 | BLKQ	   Oper1		{ipc(o1[0]<<2);uic();}
 | ADDR    Qnumber		{}
 | QUAD    Qnumber		{}
 | LONG	   Lnumber		{}   
 | ALIGN   IMMED		{apc(R[0]); nl();}
 | ALIGN   IM64			{apc(R[0]); nl();}
 | ALIGN   Oper1 CA Oper2	{apcf(o1[0],o2[0]); nl();}
 ;
Assigment:                       
   DOT  '='Assign NL	{spc(ev[1],ev[0]);endline();} 
 | SYMB			{strcpy(S1,Sym);}
	'='Assign NL	{symdef(S1,ev[1],ev[0],S2,S3,sop); endline();}
 ;

Assign:
   Oper1 '+' Oper2	{ADD64(o1,o2,ev); if(st & 1)sop=OADD;}
 | Oper1 '-' Oper2	{SUB64(o1,o2,ev); if(st & 1)sop=OSUB;}
 | Oper1 '&' Oper2	{ev[0]=o1[0] & o2[0];ev[1]=o1[1] & o2[1]; if(st & 1)sop=OAND;}
 | Oper1 OR  Oper2	{ev[0]=o1[0] | o2[0];ev[1]=o1[1] | o2[1]; if(st & 1)sop=OORR;}
 | Oper1 XOR Oper2	{ev[0]=o1[0] ^ o2[0];ev[1]=o1[1] ^ o2[1]; if(st & 1)sop=OXOR;}
 | Oper1 '@' Oper2	{ev[0]=((int)o2[0]<0) ? (o1[0]>> (-(int)(o2[0])))
					      : (o1[0]<<o2[0]);}
 | Oper1 '~'		{ev[0]= ~o1[0];ev[1]= ~o1[1];}
 | Oper1		{ev[0]=o1[0];ev[1]=o1[1]; if(st & 1)sop=OONE;}
 | error		{yyerror("No operand for the Assignment");}
 ;
Oper1:
   DOT		{o1[0]=pc; sprintf(S2,"^x%x%08x",hpc,pc);}
 | IMMED	{o1[0]=R[0];o1[1]=R[1];sprintf(S2,"^x%x%08x",R[1],R[0]);}
 | IM64		{o1[0]=R[0];o1[1]=R[1];sprintf(S2,"^x%x%08x",R[1],R[0]);}
 | SYMB		{st|=symget(Sym,R); strcpy(S2,Sym); o1[0]=R[0];o1[1]=R[1];}
 | error	{yyerror("Bad 1st Operand");}
 ;
Oper2:
   DOT		{o2[0]=pc; sprintf(S3,"^x%x%08x",hpc,pc);}
 | IMMED	{o2[0]=R[0];o2[1]=R[1];sprintf(S3,"^x%x%08x",R[1],R[0]);}
 | IM64		{o2[0]=R[0];o2[1]=R[1];sprintf(S3,"^x%x%08x",R[1],R[0]);}
 | SYMB		{st|=symget(Sym,R); strcpy(S3,Sym); o2[0]=R[0];o2[1]=R[1];}
 | error	{yyerror("Bad 2nd Operand");}
 ;
Ipr:
   IMMED		{Opc|=R[0];}
 | IM64			{Opc|=R[0];}
 | IPRS			{Opc|=R[0];}
 ;
Lnumber:        
   IMMED		{db();m[ic++]=R[0];ipc(4);}
 | IM64			{db();m[ic++]=R[0];ipc(4);}
 | SYMB	    		{symuse(Sym,TD32);db();m[ic++]=0xADD;ipc(4);}
 ;
Qnumber:
   IMMED		{db();m[ic++]=R[0];db();m[ic++]=R[1]; ipc(8);}
 | IM64			{db();m[ic++]=R[0];db();m[ic++]=R[1]; ipc(8);}
 | SYMB			{symuse(Sym,TD64);db();m[ic++]=0xADD;
					  db();m[ic++]=0xADD;ipc(8);}
 ;
Mode:
   /*empty*/	
 | CA SYMB		{symuse(Sym,TMOD);}
 | CA IMMED		{Opc |= (R[0] & 0xF) << 12;}
 | CA IM64		{Opc |= (R[0] & 0xF) << 12;}
 ;
Disp12:
			{}
 | SYMB			{symuse(Sym,TMHW);}
 | IMMED		{Opc |= R[0] & 0xFFF;}
 | IM64			{Opc |= R[0] & 0xFFF;}
 ;
Disp16:
			{}
 | SYMB			{symuse(Sym,TMDS);}
 | IMMED		{Opc |= R[0] & 0xFFFF;}
 | IM64			{Opc |= R[0] & 0xFFFF;}
 ;
Disp21:
   SYMB			{symuse(Sym,TBDS);}
 | IMMED		{Opc |= 0x1FFFFF & R[0];}
 | IM64			{Opc |= 0x1FFFFF & R[0];}
 | DOT			{Opc |= 0x1FFFFF & (0>>2);}  /*Interesting ??*/
 ;
ThreeFReg:
   REG CA		{Opc |= (Reg<<21); }
   REG CA		{Opc |= (Reg<<16); }
   REG			{Opc |= (Reg);	   }
 | REG error		{yyerror("Syntax, FRegisters?");}
 ;
IReg:
   REG CA		{Opc |= (Reg<<16); }
   REG			{Opc |= (Reg);     }
 | PD IMMED CA	{Opc |= (R[0]<<13); Opc|=0x1000;}
   REG			{Opc |= (Reg);     }
 | PD IM64 CA		{Opc |= (R[0]<<13); Opc|=0x1000;}
   REG			{Opc |= (Reg);     }
 | PD SYMB CA		{symuse(Sym,TIOP); Opc|=0x1000;}
   REG			{Opc |= (Reg);     }
 | SYMB error		{yyerror("%SYNTAX: use # for Immediate values\n");}
 | REG error		{yyerror("%SYNTAX: IRegisters?");}
 ;
TwoReg:
   REG CA		{Opc |= ((31<<21) | (Reg<<16));}
   REG			{Opc |= (Reg);	  }
 | REG error		{yyerror("Syntax, Registers?");}
 ;
%%
/*****************************************************************************/
yyerror(s)
char *s;
{
static int counter =0;
 counter++;
 printf("%%HAL ERROR: %s \n\tAt line %d While processing token=<%s>\n",
				s,line_number,yylval);
 printf("             %s\n",Sline);
 fprintf(Ot,"%%HAL ERROR: %s \n\tAt line %d While processing token=<%s>\n",
				s,line_number,yylval);
 fprintf(Ot,"             %s\n",Sline);
 Kcode=1;			/* Do not generate code on Error*/
 if(counter==5)			/* After 5 fatal error, Die!*/
  {
   generate_list(NULL);
   exit(EXIT_FAILURE);
  };
 return;
}
/*****************************************************************************/
/* Update IC from PC							     */
/*****************************************************************************/
uic()
{      
/* Original equation: ic=(pc-Saddr)>>2; */
unsigned int a[2],b[2],c[2];
 a[1]=hpc;   a[0]=pc;
 b[1]=Sa[1]; b[0]=Sa[0];
 SUB64(a,b,c);
 ic = c[0] >> 2;
}
/*****************************************************************************/
/* Increment PC								     */
/*****************************************************************************/
ipc(d)
int d;
{
unsigned int a[2],b[2],c[2];
 a[1]=hpc; a[0]=pc;
 b[1]=0x0; b[0]=d;
 ADD64(a,b,c);
 hpc=c[1];
  pc=c[0];
 endline();
}
/*****************************************************************************/
/* Align PC								     */
/*****************************************************************************/
apc(d)
unsigned int d;
{
unsigned int a[2],b[2],c[2];
 d = (1 << d) - 1;  /* 2^d - 1 */
 if(!(pc & d)) return;
 pc &= ~d;
 d++;
 a[1]=hpc; a[0]=pc;
 b[1]=0x0; b[0]=d;
 ADD64(a,b,c);
 hpc=c[1];
  pc=c[0];
 uic();
 return;
}
/*****************************************************************************/
/* Align PC with filling						     */
/*****************************************************************************/
apcf(d,fv)
B32 d, fv;
{
unsigned int a[2],b[2],c[2];
B32	sic;
 d = (1 << d) - 1;  /* 2^d - 1 */
 if(!(pc & d)) return;
 pc &= ~d;
 d++;
 a[1]=hpc; a[0]=pc;
 b[1]=0x0; b[0]=d;
 ADD64(a,b,c);
 hpc=c[1];
  pc=c[0];
 sic=ic;
 uic(); 
 for(;sic<ic;sic++) m[sic]=fv;
 return;
}
/*****************************************************************************/
/*****************************************************************************/
dbug()
{
 fprintf(Ot,"Opc=%08X\n",Opc); 
 ref[ic].g  = 1;
 ref[ic].ln = line_number;
}
/*****************************************************************************/
/*****************************************************************************/
db()
{
 ref[ic].g  = 1;
 ref[ic].ln = line_number;
}
/*****************************************************************************/
/* RESET after each line						     */
/*****************************************************************************/
rs()
{
 st=0; 
 sop = NONE;
 strcpy(S1,"____");
 strcpy(S2,"____");
 strcpy(S3,"____");
}
/*****************************************************************************/
/* Used by Parser to concatenate the tokens                                  */
/*****************************************************************************/
YYSTYPE add(s1,s2)
char *s1, *s2;
{
 char *new;
 new = (char *)malloc((strlen(s1)+strlen(s2)+2) * sizeof(char));
 sscanf(new,"%s %s",s1,s2);
 if(s1[0] == '\0' || s1 == NULL) strcpy(new,s2);
 if(s2[0] == '\0' || s2 == NULL) strcpy(new,s1);
 return(new);
}

