/*
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.

*/
#include <stdio.h>
#include <string.h>
#include <signal.h>

#define main
#include "pp.h"
#undef main

extern int yydebug;
extern int total_alloc;
extern FILE *yyout;

#ifdef __STDC__
void parse_cmdline(int argc, char *argv[])
#else
void parse_cmdline(argc, argv)
int argc;
char *argv[];
#endif
{
  FILE *tfil;
  int i;

  /* skip over the image name */
  for (i=1; i<argc; ++i) {
    char *arg = argv[i];

    switch (*arg) {
    case '-':
      switch (*(++arg)) {
      case 'o':
	++i;
	if ((tfil = fopen(argv[i], "w")) == NULL) {
	  errout(E_ERR | E_NO_LNUM, "parse_cmdline", "error opening file %s for output, using stdout.\n", argv[i]);
	} else
	  yyout = tfil;
	break;
      case 'l':
	/* listing file */
	break;
#if 0
      case 'p':
	preprocess_only = 1;
	break;
#endif
      case 'd':
	dot_to_next_inst = 0;
	break;
      case 'q':
	quiet = 1;
	break;
      case 'm':
	if (*(++arg) == '\0')
	  arg = argv[++i];
	max_macro_size = atol(arg);
	if (max_macro_size == 0) {
	  errout(E_WARN | E_NO_LNUM, "parse_cmdline", "max_macro_size must be non-zero, using 4096.\n");
	  max_macro_size = 4096;
	}
	break;
      case 'i':
	++i;
	add_input(argv[i], F_FILE, 1, ADD_APPEND);
	break;
      }
      break;
    default:
      add_input(arg, F_FILE, 1, ADD_APPEND);
    }
  }

  return;
}

void init()
{
  /* must be first, avl_init_table uses the memmgt system */
  init_memmgt();

#if 0
  preprocess_only = 0;
#endif
  dot_to_next_inst = 1;
  quiet = 0;
  max_macro_size = 4096;
  local_label = 30000;
  ifile = NULL;
  cur_ifile = NULL;
  symtree = avl_init_table(strcmp);
  symtab = avl_init_table(strcmp);
  macrotree = avl_init_table(strcmp);
  defs = NULL;
  line_size = 2048;
#if defined(VMS) || defined(sun)
  /* both VMS and sun's fail when NULL is passed to realloc
     so much for standards, huh? */
  lptr = line = (char *)malloc(line_size + 1);
#else
  lptr = line = (char *)realloc(NULL, line_size + 1);
#endif
  macstk = NULL;
  macro_temp = NULL;
  cur_psect = psect_stk = NULL;

  return;
}

void uninit()
{
  free(line);

  end_memmgt();

  return;
}

#ifdef __STDC__
main(int argc, char *argv[])
#else
main(argc, argv)
int argc;
char *argv[];
#endif
{
  int status;

  init();

  parse_cmdline(argc,argv);

#ifdef __DATE__
    errout(E_INFO | E_NO_LNUM, "pp", "HAL preprocesser %s, compiled %s %s\n",
	   VERSION, __DATE__, __TIME__);
#else
    errout(E_INFO | E_NO_LNUM, "pp", "HAL preprocesser %s\n",
	   VERSION);
#endif

  /* open first input file */
  status = next_file();

  if (!status) {
    errout(E_WARN | E_NO_LNUM, "pp", "no input files specified.\n");
    uninit();
    exit(EXIT_FAILURE);
  }

  macro_temp = ALLOCn(max_macro_size);

  oldsegv = signal(SIGSEGV, segv_catcher);

#ifdef DEBUG
  yydebug = 1;
#endif
  (void)yyparse();

  output_assignments(defs);

  if (yyout != stdout)
    fclose(yyout);
  
  uninit();

  errout(E_INFO | E_NO_LNUM, "pp", "total memory requested: %d bytes (%dK).\n", total_alloc, (total_alloc + 1023)/1024);

  (void)signal(SIGSEGV, oldsegv);

#if 0
  if (!preprocess_only) {
    /* execve to the assembler */
  }
#endif

  exit(EXIT_SUCCESS);
}
